package org.processmining.hybridminer.models.hybridpetrinet;

import org.apache.commons.math3.util.CombinatoricsUtils;
import org.deckfour.xes.model.XLog;
import org.processmining.hybridminer.models.hybridpetrinet.candidateplaceiterator.*;
import org.processmining.hybridminer.models.hybridpetrinet.candidateplacestrategy.CandidatePlaceSelectionStrategy;
import org.processmining.hybridminer.models.hybridpetrinet.candidateplacestrategy.EdgesOrderingCompatibleSelStrategy;
import org.processmining.hybridminer.models.hybridpetrinet.conflictstrategy.*;
import org.processmining.models.graphbased.NodeID;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphEdge;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphNode;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;



/**
 * Created by demas on 19/08/16.
 */

public class Cluster<E extends AbstractDirectedGraphEdge, N extends AbstractDirectedGraphNode> implements Runnable {
//public class Cluster<E extends AbstractDirectedGraphEdge, N extends AbstractDirectedGraphNode> {
	private Set<E> edges;
	private Set<N> inputNodes, outputNodes;

	//private Set<PlaceEvaluation<N>> places;

	// new variables for the thread implementation, to be initialized later.
	private XLog log;
	private Map<String, Integer> activityFrequencyMap;
	private double placeEvaluationThreshold;
	private double prePlaceEvaluationThreshold;

	// new parameters for performance optimization
	//private Set<PlaceEvaluation<N>> placesToBeAdded;
	private CandidatePlaceIterator<N> placeIterator;
	private ConflictStrategy<N> conflictStrategy;
	private volatile boolean stopped;
	private AtomicInteger maxPlaceNumber;


	public Cluster(Set<E> edges, CandidatePlaceSelectionStrategy<N> selectionStrategy, CandidatePlaceIteratorEnum placeIteratorType, ConflictStrategyEnum conflictStrategyType, AtomicInteger maxPlaceNumber) {
		this.edges = edges;
		this.inputNodes = new HashSet<>();
		this.outputNodes = new HashSet<>();
		computeInputAndOutputNodes();
		this.conflictStrategy = selectConflictStrategy(conflictStrategyType);
		this.stopped = false;
		this.maxPlaceNumber = maxPlaceNumber;
		this.placeIterator = selectPlaceIterator(selectionStrategy, placeIteratorType);
	}

	private ConflictStrategy<N> selectConflictStrategy(ConflictStrategyEnum conflictStrategyType) {
		switch(conflictStrategyType) {
			case DEFAULT_CONFLICT_STRATEGY:
				return new DefaultConflictStrategyNonOpt<N>();

			case DEFAULT_CONFLICT_STRATEGY_OPT:
				return new DefaultConflictStrategyOpt<N>();

			case NO_CONFLICT_STRATEGY:
				return new NoConflictStrategy<N>();

			default:
				throw new RuntimeException("Please select a valid conflict strategy!");
		}
	}



	public void printDotFile() {
		/*
        Printing to .gv (graphviz) file
		 */
		FileOutputStream fos = null;

		try {
			fos = new FileOutputStream(new Integer(System.identityHashCode(this)).toString() + ".gv");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		PrintStream ps = new PrintStream(fos);
		ps.println(this.toDot());
		ps.flush();
		ps.close();
	}


	public String toDot() {
		Map<NodeID, Integer> nodeNameMap = new HashMap<>();

		StringBuilder b = new StringBuilder("digraph Cluster {\n");
		b.append("  rankdir = LR;\n");
		Set<N> IONodes = new HashSet<>();
		IONodes.addAll(this.outputNodes);
		IONodes.addAll(this.inputNodes);

		Set<N> intersectionIONodes = new HashSet<>();
		intersectionIONodes.addAll(this.outputNodes);
		intersectionIONodes.retainAll(this.inputNodes);
		System.out.println("Cluster: " + System.identityHashCode(this) + " has only " + (IONodes.size() - intersectionIONodes.size()) + " which are both input and output." );


		// print states;
		int counter = 0;
		for (N n : IONodes) {
			Integer newNodeName = nodeNameMap.get(n.getId());
			if (newNodeName==null) {
				newNodeName = new Integer(counter);
				nodeNameMap.put(n.getId(), newNodeName);
				counter++;
			}
			b.append("  ").append(newNodeName.toString());
			b.append(" [shape=circle,label=\"" + newNodeName.toString() + "\"];\n");
		}

		for (E e : this.edges) {
			NodeID idSource = ((N) e.getSource()).getId();
			NodeID idTarget = ((N) e.getTarget()).getId();
			b.append("  ").append(nodeNameMap.get(idSource).toString());
			b.append(" -> ").append(nodeNameMap.get(idTarget).toString()).append(" [label=\"");
			b.append("\"]\n");
		}
		return b.append("}\n").toString();
	}


	private CandidatePlaceIterator<N> selectPlaceIterator(CandidatePlaceSelectionStrategy<N> selectionStrategy, CandidatePlaceIteratorEnum placeIteratorType) {
		switch(placeIteratorType) {
			case OUTPUT_THEN_INPUT_NUMBER :
				return new OutputThenInputOrderingIterator<N>(selectionStrategy, this.outputNodes, this.inputNodes);

			case EDGE_NUMBER :
				return new EdgesOrderingIterator<N, E>(selectionStrategy, this.edges);

			case EDGE_NUMBER_OPT :
				if (! (selectionStrategy instanceof EdgesOrderingCompatibleSelStrategy))
					throw new RuntimeException("In order to use the EDGE_NUMBER_OPT ordering strategy, the set of candidate place must be compatible, i.e., either BOUNDED_INPUT_PLUS_OUTPUT or SPLIT_JOIN_ONLY_BOUND!");
				return new EdgesOrderingIteratorOpt<N, E>((EdgesOrderingCompatibleSelStrategy<N>)selectionStrategy, this.edges);

			default:
				throw new RuntimeException("Please select or implement the right CandidatePlaceIterator class!");
		}
	}


	private void computeInputAndOutputNodes() {
		for(E edge : this.edges) {
			this.inputNodes.add((N) edge.getTarget());
			this.outputNodes.add((N) edge.getSource());
		}
	}


	public void setPrePlaceEvaluationThreshold(double prePlaceEvaluationThreshold) {
		this.prePlaceEvaluationThreshold = prePlaceEvaluationThreshold;
	}

	public void setPlaceEvaluationThreshold(double placeEvaluationThreshold) {
		this.placeEvaluationThreshold = placeEvaluationThreshold;
	}

	public Set<E> getEdges() {
		return new HashSet<>(edges);
	}


	public Set<N> getInputNodes() {
		return new HashSet<>(inputNodes);
	}


	public Set<N> getOutputNodes() {
		return new HashSet<>(outputNodes);
	}


	public void setLog(XLog log) {
		this.log = log;
	}

	public void setActivityFrequencyMap(Map<String, Integer> activityFrequencyMap) {
		this.activityFrequencyMap = activityFrequencyMap;
	}


	public Set<PlaceEvaluation<N>> getPlacesToBeAdded() {
		return this.conflictStrategy.getPlacesToBeAdded();
	}


	public void run() {
		int originalNumPlacesBound = this.maxPlaceNumber.get();
		long startTime, endTime, elapsedTimeSec;
		startTime = System.currentTimeMillis();

		while(!this.stopped && !maxPlaceNumber.compareAndSet(0, 0) && this.placeIterator.hasNext()) {
			endTime = System.currentTimeMillis();
			elapsedTimeSec = (endTime - startTime)/1000;
			System.out.println(" END iterator.hasNext() in " + elapsedTimeSec + "sec.");


			System.out.print("START iterator.next()... ");
			startTime = System.currentTimeMillis();
			PartialPlaceEvaluation<N> candidatePPE = this.placeIterator.next();
			endTime = System.currentTimeMillis();
			elapsedTimeSec = (endTime - startTime)/1000;
			System.out.println(" END iterator.next() in " + elapsedTimeSec + "sec.");

			
			if (!candidatePPE.isSyphon() && checkPlaceConsistency(candidatePPE, this.edges)) {

				System.out.print("START checkConflict... ");
				startTime = System.currentTimeMillis();
				
				if (!checkConflict(candidatePPE)) {
					endTime = System.currentTimeMillis();
					elapsedTimeSec = (endTime - startTime)/1000;
					System.out.println(" END checkConflict in " + elapsedTimeSec + "sec.");

					PlaceEvaluation<N> candidatePlaceEval = new PlaceEvaluation<N>(candidatePPE, log, activityFrequencyMap, prePlaceEvaluationThreshold, placeEvaluationThreshold);
					System.out.print("START evaluating on log PE with " + candidatePlaceEval.getPlaceOutputNodes().size() + " input nodes and " + candidatePlaceEval.getPlaceInputNodes().size() + " output nodes... ");
					startTime = System.currentTimeMillis();
					candidatePlaceEval.replayPlace();
					endTime = System.currentTimeMillis();
					elapsedTimeSec = (endTime - startTime)/1000;
					System.out.println("END evaluation on log in " + elapsedTimeSec + "sec.");

					System.out.print("START addPlace through ConflictStrategy... ");
					startTime = System.currentTimeMillis();
					if (candidatePlaceEval.evaluateReplayScore() >= placeEvaluationThreshold) {
						//synchronized(maxPlaceNumber) {
						addPlace(candidatePlaceEval);
						maxPlaceNumber.getAndDecrement();
						System.out.print("Place added! Number of places: " + (originalNumPlacesBound-this.maxPlaceNumber.get()));
						//}
						//System.out.println("Place added: " + candidatePlaceEval);
						endTime = System.currentTimeMillis();
						elapsedTimeSec = (endTime - startTime)/1000;
						System.out.println(" END addPlace " + elapsedTimeSec + "sec.");
					}
				}
				else {
					endTime = System.currentTimeMillis();
					elapsedTimeSec = (endTime - startTime)/1000;
					System.out.println(" END checkConflict in " + elapsedTimeSec + "sec.");
				}
			}

			System.out.print("START iterator.hasNext()... ");
			startTime = System.currentTimeMillis();	
		}
		endTime = System.currentTimeMillis();
		elapsedTimeSec = (endTime - startTime)/1000;
		System.out.println(" END iterator.hasNext() in " + elapsedTimeSec + "sec.");
	}


	private void addPlace(PlaceEvaluation<N> p) {
		this.conflictStrategy.addPlace(p);
	}


	public void stop() {
		this.stopped=true;
	}


	/*
	 * Checks if ppe is in conflict with set places according to the selected strategy.
	 */
	private <T extends PartialPlaceEvaluation<N>> boolean checkConflict(PartialPlaceEvaluation<N> ppe) {
		//		long startTime = System.currentTimeMillis();
		boolean result = this.conflictStrategy.checkConflict(ppe);
		//		long stopTime = System.currentTimeMillis();
		//		System.out.println("CheckConflict required: " + (stopTime-startTime) + "ms" + " conflict found: " + result);
		return result;
	}




	/*
	 * Given a place p, check if **each** output node there exists an edge in the graph with **ALL** input nodes.
	 */
	public static <E extends AbstractDirectedGraphEdge, N extends AbstractDirectedGraphNode> boolean checkPlaceConsistency(PartialPlaceEvaluation<N> ppe, Set<E> edges) {
		Set<N> outputNodes = ppe.getPlaceOutputNodes();
		Set<N> inputNodes = ppe.getPlaceInputNodes();
		boolean result = true;
		for (N outNode : outputNodes) {
			for (N inNode : inputNodes) {
				if (!containsEdge(outNode, inNode, edges)) {
					//System.out.println("PlaceEvaluation not consistent! OutputNodes: " + outNode + " InputNodes: " + inNode);
					return false; 
				}
			}
		}
		return result;
	}


	private static <E extends AbstractDirectedGraphEdge, N extends AbstractDirectedGraphNode> boolean containsEdge(N outputNode, N inputNode, Set<E> edges) {
		for (E edge : edges) {
			if (edge.getSource().equals(outputNode) && edge.getTarget().equals(inputNode))
				return true;
		}
		return false;
	}




	/*public Set<PlaceEvaluation<N>> getNonRedundantPlacesAboveThreshold(double threshold) {
        Set<PlaceEvaluation<N>> aboveThreshold = this.getPlacesAboveThreshold(threshold);
        //aboveThreshold.removeAll(getRedundantPlaces(aboveThreshold));
        return aboveThreshold;
    }*/


	@Override
	public String toString() {
		return "Cluster{" +
				"edges=" + edges +
				", inputNodes=" + inputNodes +
				", outputNodes=" + outputNodes +
				'}';
	}


	public String printStatistics() {
		StringBuffer result = new StringBuffer();
		result.append("*** Info for cluster " + System.identityHashCode(this) + " ***\n");
		result.append("No. of output nodes: " + this.getOutputNodes().size() + "\n");
		result.append("No. of input nodes: " + this.getInputNodes().size() + "\n");
		result.append("No. of edges: " + this.getEdges().size() + "\n");

		int maxInput = 0;
		int maxOutput = 0;
		for (N n : this.getOutputNodes()) {
			int maxNodeOutput = 0;
			int maxNodeInput = 0;
			for (E e : this.getEdges()) {
				if (e.getSource().equals(n))
					maxNodeOutput++;
				if (e.getTarget().equals(n))
					maxNodeInput++;
			}
			if (maxNodeOutput > maxOutput)
				maxOutput = maxNodeOutput;
			if (maxNodeInput > maxInput)
				maxInput = maxNodeInput;
		}

		for (N n : this.getInputNodes()) {
			int maxNodeOutput = 0;
			int maxNodeInput = 0;
			for (E e : this.getEdges()) {
				if (e.getSource().equals(n))
					maxNodeOutput++;
				if (e.getTarget().equals(n))
					maxNodeInput++;
			}
			if (maxNodeOutput > maxOutput)
				maxOutput = maxNodeOutput;
			if (maxNodeInput > maxInput)
				maxInput = maxNodeInput;
		}

		result.append("Max outgoing edges: " + maxOutput + ". Max ingoing edges " + maxInput + "\n");

		int maxIO = Math.max(maxOutput, maxInput);

		double candidateNum = 0;
		for (int i=1; i<=maxIO; i++)
			candidateNum = candidateNum + CombinatoricsUtils.binomialCoefficientDouble(this.getEdges().size(), i);

		result.append("No. of possible candidates (not consistent are included): " + candidateNum + "\n");

		return new String(result);
	}


	/*    private static <N extends AbstractDirectedGraphNode> Set<PlaceEvaluation<N>> getRedundantPlaces(Set<PlaceEvaluation<N>> placesAboveThreshold) {
        System.out.println("Redundant place computation started!");
    	Set<PlaceEvaluation<N>> toBeDiscarded = new HashSet<>();

        for (PlaceEvaluation<N> p : placesAboveThreshold) {
            if(isRedundant(placesAboveThreshold, p))
                toBeDiscarded.add(p);
        }
        System.out.println("Redundant place computation ended!");
        return toBeDiscarded;
    }*/


	/* To be called AFTER evaluateBestPlaces!
       The method scans this.places and sees if pe is redundant, i.e., if there is a set of placeEvaluation such that:
       the union of input sets is equal to pe input set
       the intersection of input sets is empty
       the union of output sets is equal to pe output set
       the intersection of output sets is empty
	 */
	/*private static <N extends AbstractDirectedGraphNode> boolean isRedundant(Set<PlaceEvaluation<N>> aboveThreshold, PlaceEvaluation<N> pe) {
        // First of all, remove pe from above threshold.
        Set<PlaceEvaluation<N>> otherPlaces = new HashSet<>();
        otherPlaces.addAll(aboveThreshold);
        if (!otherPlaces.remove(pe))
            throw new RuntimeException("The input place evaluation should be contained in the set of place evaluations above the threshold");

        // We need the array to access elements by index (the apache math methods returns an iterator on indices)
        PlaceEvaluation<N>[] otherPlacesArray = (PlaceEvaluation<N>[]) otherPlaces.toArray(new PlaceEvaluation<?>[otherPlaces.size()]);


        // We need to try every possible combination of i elements from the set otherPlacesArray.
        for (int i=2; i < otherPlacesArray.length; i++) {

            Iterator<int[]> it = org.apache.commons.math3.util.CombinatoricsUtils.combinationsIterator(otherPlacesArray.length, i);

            // Each next() is an array of indices representing the current choice of elements.
            while (it.hasNext()) {
                int[] indexesArray = it.next();
                //System.out.println("OtherPlaces number: " + otherPlacesArray.length + ". Current combination: " + Arrays.toString(indexesArray));
                Set<PlaceEvaluation<N>> currentCombination = new HashSet<>();

                // Build the set of PlacesEvalutions corresponding to the current choice of elements
                for (int j=0; j<indexesArray.length; j++)
                    currentCombination.add(otherPlacesArray[indexesArray[j]]);

                // Call the method that actually checks the redundancy
                if (checkRedundancy(currentCombination, pe))
                    return true;
            }
        }
        return false;
    }


    private static <N extends AbstractDirectedGraphNode> boolean checkRedundancy(Set<PlaceEvaluation<N>> currentCombination, PlaceEvaluation<N> placeEval) {
        // Check nullity of inputs!
        if (currentCombination==null || currentCombination.size()<2 || placeEval==null)
            throw new RuntimeException("Check the inputs of checkRedundancy!");

        // Build the union and the intersection of inputPlaces and outputPlaces of the currentCombination
        Set<N> intersectionInputPlaces = new HashSet<>();
        Set<N> intersectionOutputPlaces = new HashSet<>();
        Set<N> unionInputPlaces = new HashSet<>();
        Set<N> unionOutputPlaces = new HashSet<>();

        PlaceEvaluation<N>[] currentCombinationArray = (PlaceEvaluation<N>[]) currentCombination.toArray(new PlaceEvaluation<?>[currentCombination.size()]);
        intersectionInputPlaces.addAll(currentCombinationArray[0].getPlaceInputNodes());
        intersectionOutputPlaces.addAll(currentCombinationArray[0].getPlaceOutputNodes());

        for (PlaceEvaluation<N> p : currentCombination) {
            intersectionInputPlaces.retainAll(p.getPlaceInputNodes());
            intersectionOutputPlaces.retainAll(p.getPlaceOutputNodes());
            unionInputPlaces.addAll(p.getPlaceInputNodes());
            unionOutputPlaces.addAll(p.getPlaceOutputNodes());
        }

        if (intersectionInputPlaces.size() != 0 || intersectionOutputPlaces.size() != 0)
            return false;

        if (unionInputPlaces.equals(placeEval.getPlaceInputNodes()) && unionOutputPlaces.equals(placeEval.getPlaceOutputNodes())){
        	System.out.println("Removed "+placeEval+" because of:");
        	for (PlaceEvaluation<N> placeEvaluation : currentCombinationArray) {
				System.out.println(placeEvaluation);
			}
        }

        return (unionInputPlaces.equals(placeEval.getPlaceInputNodes()) && unionOutputPlaces.equals(placeEval.getPlaceOutputNodes()));
    }*/


}
