package org.processmining.hybridminer.models.hybridpetrinet;

import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension.StandardModel;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphNode;

/**
 * Created by demas on 22/08/16.
 */
public class PlaceEvaluation<N extends AbstractDirectedGraphNode> extends PartialPlaceEvaluation<N> {
    private int acceptedTracesNumber;
    private int currentTokenNumber;
    private XLog log;
    private Map<String, Integer> activityFrequencyMap;
    double prePlaceEvaluationThreshold;
    double placeEvaluationThreshold;
    // A trace in the log is "active" if, during the replay it has activated at least one outputTransition or inputTransition.
    int activeTracesNumber;
    
    
    public PlaceEvaluation(Set<N> placeOutputNodes, Set<N> placeInputNodes, XLog log, Map<String, Integer> activityFrequencyMap, double prePlaceEvaluationThreshold, double placeEvaluationThreshold) {
        super(placeOutputNodes, placeInputNodes);
        this.log = log;
        this.acceptedTracesNumber = 0;
        this.currentTokenNumber = 0;
        this.activityFrequencyMap = activityFrequencyMap;
        this.prePlaceEvaluationThreshold = prePlaceEvaluationThreshold;
        this.placeEvaluationThreshold = placeEvaluationThreshold;
        this.activeTracesNumber = 0;
    }
    
    
    public PlaceEvaluation(PartialPlaceEvaluation<N> ppe, XLog log, Map<String, Integer> activityFrequencyMap, double prePlaceEvaluationThreshold, double placeEvaluationThreshold) {
    	this(ppe.getPlaceOutputNodes(), ppe.getPlaceInputNodes(), log, activityFrequencyMap, prePlaceEvaluationThreshold, placeEvaluationThreshold);
    }


	@Override
    public String toString() {
        return "PlaceEvaluation{" +
                "placeInputNodes=" + this.getPlaceInputNodes() +
                ", placeOutputNodes=" + this.getPlaceOutputNodes() +
                '}';
    }

  

	/*
     * This methods is for optimizing large cluster. It pre-evaluates a specific inputSet and outputSet set and analyzes if
     * it is a good candidate if:
     * 				AbsoluteValue[Sum_i(InputSet(i)) - Sum_j(OutputSet(j))] / Sum_i(InputSet(i)) + Sum_j(OutputSet(j))  <  prePlaceEvaluationThreshold
     */
    public boolean preEvaluate() {
    	double inputSum = 0;
    	double outputSum = 0;
    	for (N node : this.getPlaceInputNodes())
    		inputSum += this.activityFrequencyMap.get(node.getLabel());
    	for (N node : this.getPlaceOutputNodes())
    		outputSum += this.activityFrequencyMap.get(node.getLabel());
    	double divisor = Math.abs(inputSum - outputSum);
    	double dividend = inputSum + outputSum;
    	//System.out.println("Divisor/dividend:" + divisor/dividend);
    	return ((divisor/dividend) < this.prePlaceEvaluationThreshold);
    }
    

    public void replayPlace() {
    	double numAnalyzedTraces = new Double(0);
    	
        // for each trace in the log
    	for (XTrace trace : log) {
    		//this.outputOccurred = false;
    		boolean peActivatedByTrace;
    		peActivatedByTrace = this.replayPlaceOnTrace(trace);
    		if (peActivatedByTrace) {
    			this.activeTracesNumber++;
    		}
            // update Accepted traces
    		//if (isCurrentTokenNumberZero() || this.outputOccurred==false)
    		if (isCurrentTokenNumberZero() && peActivatedByTrace)
    			increaseAcceptedTracesNumber();
    		resetCurrentTokenNumber();
    		
    		
    		// OPTIMIZATION in appendix of paper v4.
    		numAnalyzedTraces++;
    		if ((runtimePlaceEvaluation(numAnalyzedTraces) < this.placeEvaluationThreshold)) {
    			//System.out.println("Evaluation of a log interrupted for optimizing.");
    			return;
    		}
    		
    		/* Notice that when returns, the evaluateReplayScore() still is less than placeEvaluationThreshold, for the theorem
    		 (a + x) / (b + x) < c ===> a/b < c when a>b, which is always the case.
    		 */
    	}
    }
    
    
    // As in appendix of the paper in v4
    double runtimePlaceEvaluation(double numAnalyzedTraces) {
    	double logLength = new Double(log.size());
    	double numRemainingTraces = logLength - numAnalyzedTraces;
    	
    	double numerator = getAcceptedTracesNumber() + numRemainingTraces;
    	double denominator = activeTracesNumber + numRemainingTraces;
    	
    	return numerator/denominator;
    }
    
    
    /*
     * The return value is true if trace "activated" the place evaluation, i.e., the trace had at least one
     * event corresponding to an output transition (so input to the place) or an input transition; otherwise it returns false.
     */
    public boolean replayPlaceOnTrace(XTrace trace){
    	boolean result = false;
    	
		for (XEvent event : trace) {
			String eventName = XConceptExtension.instance().extractName(event);
			StandardModel eventTransition = XLifecycleExtension.instance().extractStandardTransition(event);
			if (eventTransition==null || eventTransition.equals(XLifecycleExtension.StandardModel.COMPLETE)){
			
				boolean isInput=false, isOutput = false;
				for (N placeOutputNode : getPlaceOutputNodes()) {
					if (placeOutputNode.getLabel().equalsIgnoreCase(eventName)) {
						isOutput = true;
						result = true;
					}
				}
				for (N placeInputNode : getPlaceInputNodes()) {
					if (placeInputNode.getLabel().equalsIgnoreCase(eventName)) {
						isInput=true;
						result = true;
					}
				}
				//if it is both input and output: be conservative, i.e., 
				//if we have at least a token, we decrease the token number and return it
				// otherwise, we increase the token number
//				if (isInput && isOutput){
//					if (this.currentTokenNumber>0)
//						decreaseTokenNumber();
//					else 
//						increaseTokenNumber();
//				} else{
				
					//if it is output increase
					if (isInput){
						decreaseTokenNumber();
				        // CHECK: if negative return;
						if (isCurrentTokenNumberNegative())
							return result;
					}
					if (isOutput)
						increaseTokenNumber();
					 // if it is input decrease
				//}
			}
			
		}
		return result;
    }

    public void increaseAcceptedTracesNumber() {
        acceptedTracesNumber++;
    }

    public int getAcceptedTracesNumber() {
        return acceptedTracesNumber;
    }

    /*
     * Now changed: the score is not anymore: #acceptedTraces/#allTracesInTheLog but
     * it is #acceptedTraces/#activeTraces
     */
    public double evaluateReplayScore() {
    	if (this.activeTracesNumber==0)
    		return 0.0;
    	
        double acceptedTraces = new Double(this.getAcceptedTracesNumber());
        //double logS = new Double(log.size());
    	//return acceptedTraces/logS;
        double activeTraces = new Double(this.activeTracesNumber);
        return acceptedTraces/activeTraces;
    }

    public void increaseTokenNumber() {
    	//this.outputOccurred = true;
        this.currentTokenNumber++;
    }

    public void decreaseTokenNumber() {
        this.currentTokenNumber--;
    }
    
    public boolean isCurrentTokenNumberNegative(){
    	return currentTokenNumber<0;
    }
    
    public boolean isCurrentTokenNumberZero(){
    	return currentTokenNumber==0;
    }
    
    public void resetCurrentTokenNumber(){
    	currentTokenNumber = 0;
    }
    
    public int getActiveTracesNumber() {
    	return this.activeTracesNumber;
    }

}
