package org.processmining.hybridminer.plugins;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.hybridminer.algorithms.cg2hpn.CGToHybridPN;
import org.processmining.hybridminer.dialogs.HybridPNDialog;
import org.processmining.hybridminer.models.causalgraph.HybridCausalGraph;
import org.processmining.hybridminer.models.hybridpetrinet.HybridPetrinet;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

/**
 * Created by demas on 25/07/16.
 */

@Plugin(name = "Hybrid Petri Net Miner", parameterLabels = {"Causal Graph", "Hybrid Petri Net Configuration" }, 
	    returnLabels = { "Hybrid Petri Net"}, returnTypes = {HybridPetrinet.class})
public class HybridPNMinerPlugin {

	private HybridPetrinet privateFPNMinerPlugin(PluginContext context, HybridCausalGraph fCG, HybridPNMinerSettings settings) {	
		//double t1 = System.currentTimeMillis();
        HybridPetrinet fPN = CGToHybridPN.fuzzyCGToFuzzyPN(fCG, fCG.getLog(), settings);
        fPN = addPlacesAndMarkings(context, fPN);
/*        double t2 = System.currentTimeMillis();
		System.out.println("*****************************************");
		System.out.println("TIME REQUIRED FOR FPN "+(t2 - t1));
		System.out.println("*****************************************");	*/	
		return fPN;

	}
	
	/**
	 * Add initial and end place, as well as the initial and final marking
	 * @param context
	 * @param fPN
	 * @return fPN
	 */
	private HybridPetrinet addPlacesAndMarkings(PluginContext context, HybridPetrinet fPN){
        Place startPlace = fPN.addPlace("start");
        Marking im = new Marking();
        im.add(startPlace);
        Transition startTransition = fPN.getTransition("start");
        fPN.addArc(startPlace, startTransition);
        
        Place endPlace = fPN.addPlace("end");
        Marking fm = new Marking();
        fm.add(endPlace);        
        Transition endTransition = fPN.getTransition("end");
        fPN.addArc(endTransition, endPlace);
               
        context.getProvidedObjectManager().createProvidedObject(
                "Initial marking for " + fPN.getLabel(),
                im, Marking.class, context);
        context.addConnection(new InitialMarkingConnection(fPN, 
                im));

        context.getProvidedObjectManager().createProvidedObject(
                    "Final marking for " + fPN.getLabel(),
                    fm, Marking.class, context);
        context.addConnection(new FinalMarkingConnection(fPN, fm));
        
        
		return fPN;
	}
	
		
	/**
	 * The plug-in variant that runs in any context and requires a configuration.
	 */ 
	@UITopiaVariant(affiliation = "FBK", author = "R. De Masellis et al.", email = "r.demasellis|dfmchiara@fbk.eu")
	@PluginVariant(variantLabel = "HybridPNMiner, parameters", requiredParameterLabels = { 0, 1})
	public HybridPetrinet configuredFPNMinerPlugin(PluginContext context, HybridCausalGraph fCG, HybridPNMinerSettings settings) {
		return privateFPNMinerPlugin(context, fCG, settings);
	}
	
	/**
	 * The plug-in variant that runs in any context and uses the default configuration.
	 */
	@UITopiaVariant(affiliation = "FBK", author = "R. De Masellis et al.", email = "r.demasellis|dfmchiara@fbk.eu")
	@PluginVariant(variantLabel = "HybridPNMiner, parameters", requiredParameterLabels = { 0 })
	public HybridPetrinet defaultFPNMinerPlugin(PluginContext context, HybridCausalGraph fCG) {
		// Get the default configuration.
			
		HybridPNMinerSettings settings = new HybridPNMinerSettings();
		// Do the heavy lifting.
	    return privateFPNMinerPlugin(context, fCG, settings);
	}
	
	/**
	 * The plug-in variant that runs in a UI context and uses a dialog to get the configuration.
	 */
	@UITopiaVariant(affiliation = "FBK", author = "R. De Masellis et al.", email = "r.demasellis|dfmchiara@fbk.eu")
	@PluginVariant(variantLabel = "HybridPNMiner, dialog", requiredParameterLabels = { 0})
	public HybridPetrinet dialogFPNMinerPlugin(UIPluginContext context, HybridCausalGraph fCG) {
		// Get the default configuration.
	    HybridPNMinerSettings settings = new HybridPNMinerSettings();
	     
	    // Get a dialog for this configuration.
	    HybridPNDialog dialog = new HybridPNDialog(context, fCG, settings);
	    // Show the dialog. User can now change the configuration.
	    InteractionResult result = context.showWizard("Hybrid Petri Net Settings", true, true, dialog);
	    // User has close the dialog.
	    if (result == InteractionResult.FINISHED) {
			// Do the heavy lifting.
	    	return privateFPNMinerPlugin(context, fCG, settings);
	    }
	    // Dialog got canceled.
	    return null;
	}	
	
	
}
